from pg_data_etl import helpers


def create_database(self) -> None:
    """
    - Create the database if it doesn't exist yet, via `psql`
    """

    if not self.exists():

        db_name = self.connection_params["db_name"]

        # Create the database
        commands = [
            f'{self.cmd.psql} -c "CREATE DATABASE {db_name};" {self.uri_superuser}',
            f'{self.cmd.psql} -c "CREATE EXTENSION postgis;" {self.uri}',
        ]

        for cmd in commands:
            helpers.run_command_in_shell(cmd)


def drop_database(self) -> None:
    """
    - Drop the database if it exists, via `psql`
    """

    if self.exists():
        db_name = self.connection_params["db_name"]

        command = f'{self.cmd.psql} -c "DROP DATABASE {db_name};" {self.uri_superuser}'
        helpers.run_command_in_shell(command)


def admin(self, admin_action: str) -> None:
    """
    - Allow user to `"CREATE"` or `"DROP"` the database
    """
    admin_action = admin_action.upper()

    # Check that admin_action is allowed
    options = ["CREATE", "DROP"]
    if admin_action not in options:
        print(
            f"{admin_action=} is not supported\nAvailable administration options include: {options}"
        )
        return None

    if admin_action == "CREATE":
        create_database(self)

    if admin_action == "DROP":
        drop_database(self)


def schema_add(self, schema: str) -> None:
    """
    - Create a schema if it does not yet exist
    """

    if self._can_create_schemas:

        query = f"CREATE SCHEMA IF NOT EXISTS {schema};"

        self.execute(query)



from __future__ import annotations
import configparser
from pathlib import Path

from .make_config_file import make_config_file, DB_CONFIG_FILEPATH


def read_config_file(filepath: Path | str = DB_CONFIG_FILEPATH) -> dict:
    """
    - Read the configuration file with database info, and return as a dictionary
    keyed on each entry in the file (could be one or multiple entries in the .cfg file)

    Arguments:
        filepath (Path | str): a Path or filepath as string to the configuration file.
                               Defaults to `DB_CONFIG_FILEPATH`

    Returns:
        dict: with all of the info keyed by database cluster
    """

    config = configparser.ConfigParser()
    config.read(filepath)

    all_hosts = {}

    for host in config.sections():
        all_hosts[host] = {}

        for key in config[host]:
            value = config[host][key]

            all_hosts[host][key] = value

    return all_hosts


def configurations(filepath: Path | str = DB_CONFIG_FILEPATH, verbose: bool = False) -> dict:
    """
    - Load the configuration data from file
    - If the file does not exist yet, generate the template into the filepath provided by the user

    Arguments:
        filepath (Path | str): a Path or filepath as string to the configuration file
        verbose (bool): flag that will print out the path of the configuration file used

    Returns:
        dict: with all connection info keyed by cluster
    """
    if not filepath.exists():
        make_config_file(filepath)

    if verbose:
        print(f"Loading db configurations from {filepath}")

    return read_config_file(filepath)



from __future__ import annotations

from pandas import DataFrame
from geopandas import GeoDataFrame


def sanitize_df_for_sql(df: DataFrame | GeoDataFrame) -> DataFrame | GeoDataFrame:
    """
    Clean up a dataframe column names so it imports into SQL properly.

    This includes:
        - spaces in column names replaced with underscore
        - all column names are 100% lowercase
        - funky characters are stripped out of column names

    TODO: docstring
    """

    bad_characters = [".", "-", "(", ")", "+", ":", "$", r"/"]

    # Replace "Column Name" with "column_name"
    df.columns = df.columns.str.replace(" ", "_")
    df.columns = [x.lower() for x in df.columns]

    # Remove '.' and '-' from column names.
    # i.e. 'geo.display-label' becomes 'geodisplaylabel'
    for s in bad_characters:
        df.columns = df.columns.str.replace(s, "", regex=False)

    return df


def convert_full_tablename_to_parts(tablename: str) -> tuple:
    """
    Take in a table name and return a tuple with (schema, name)

    e.g.  'my_schema.my_table'  -> ('my_schema', 'my_table')
          'my_table'            -> ('public', 'my_table')

    Arguments:
        tablename (str): name of the table, with or without a schema

    Returns:
        tuple: with two values - (schema, table)
    """
    if "." not in tablename:
        schema = "public"
        tbl = tablename
    else:
        schema, tbl = tablename.split(".")

    return (schema, tbl)


def this_is_raw_sql(table_or_sql: str) -> bool:
    """
    - Determine if a text string is a query or a table name
    - This involves converting the text to lower case and seeing
    if 'select' and 'from' both appear in the text

    Arguments:
        table_or_sql (str): text to test

    Returns:
        bool: True or False, depending on if it's a query
    """
    text = table_or_sql.lower()

    return "select" in text and "from" in text



from __future__ import annotations
from pathlib import Path

from pg_data_etl import helpers


def export_shp_with_pgsql2shp(self, table_or_sql: str, filepath: Path) -> None:
    """
    Use pgsql2shp to export a shapefile from the database.

    Valid arguments for `table_or_sql` are the name of a table or a full query.
    e.g. "pa.centerlines"
            "SELECT * FROM pa.centerlines WHERE some_column = 'some value'"
    """

    print("WARNING! pgsql2shp creates shapefiles that do not contain EPSG values")
    print("As an alternative that preserves EPSG values, use either:")
    print("\t -> Database.export_gis(method='geopandas')")
    print("\t -> Database.export_gis(method='ogr2ogr')")

    if helpers.this_is_raw_sql(table_or_sql):
        query = table_or_sql
    else:
        query = f"SELECT * FROM {table_or_sql}"

    params = self.connection_params()

    command = f'pgsql2shp -f "{filepath}" -h {params["host"]} -u {params["un"]} -P {params["pw"]} -p {params["port"]} {params["db_name"]} "{query}" '
    print(command)

    helpers.run_command_in_shell(command)


def export_shp_with_ogr2ogr(
    self, table_or_sql: str, filepath: Path, filetype: str = "ESRI Shapefile"
) -> None:
    """
    Use ogr2ogr to export a shapefile from the database.

    Valid arguments for `table_or_sql` are the name of a table or a full query.
    e.g. "pa.centerlines"
            "SELECT * FROM pa.centerlines WHERE some_column = 'some value'"
    """

    params = self.connection_params

    pg_params_for_ogr = f'PG:"host={params["host"]} user={params["un"]} password={params["pw"]} port={params["port"]} dbname={params["db_name"]}"'

    cmd = f'{self.cmd.ogr2ogr} -f "{filetype}" "{filepath}" {pg_params_for_ogr} '

    # If a query is passed, append cmd with ' -sql QUERY'
    if helpers.this_is_raw_sql(table_or_sql):
        sql = table_or_sql
        cmd += f' -sql "{sql}"'
    # Otherwise, just append the tablename to the cmd
    else:
        tablename = table_or_sql
        cmd += f" {tablename}"

    print(cmd)
    helpers.run_command_in_shell(cmd)


def export_gis_with_geopandas(
    self,
    table_or_sql: str,
    filepath: Path | str,
    filetype: str = "geojson",
    geom_col: str = "geom",
) -> None:
    """
    - Use `geopandas` to extract data from SQL and write to `.geojson` or `.shp`
    """

    filepath = Path(filepath)

    # Exit early if arguments don't match
    if filetype not in filepath.suffix:
        print("File type and path do not match!")
        print(f"{filetype=} {filepath.suffix=}")
        return None

    if filetype not in ["geojson", "shp"]:
        print(f"Invalid filetype: {filetype=}")
        return None

    # Get a geodataframe from SQL
    if helpers.this_is_raw_sql(table_or_sql):
        query = table_or_sql
    else:
        query = f"SELECT * FROM {table_or_sql}"

    data = self.gdf(query, geom_col=geom_col)

    # Write to file
    if filetype == "geojson":
        data.to_file(filepath, driver="GeoJSON")

    elif filetype == "shp":
        data.to_file(filepath, driver="ESRI Shapefile")

    return None


def export_gis(self, method="geopandas", **kwargs):
    """
    - All methods require kwargs `table_or_sql` and `filepath`
    - Optional kwargs include `filetype` (ogr2ogr & geopandas) and `geom_col` (geopandas only)
    """
    method_mapper = {
        "geopandas": export_gis_with_geopandas,
        "ogr2ogr": export_shp_with_ogr2ogr,
        "pgsql2shp": export_shp_with_pgsql2shp,
    }

    if method not in method_mapper:
        print(f"{method=} does not exist. Valid options include: {method_mapper.keys()}")

    func = method_mapper[method]

    func(self, **kwargs)



import pandas as pd
import sqlalchemy


def df(self, query: str) -> pd.DataFrame:
    """
    - Return a `pandas.Dataframe` from a SQL query

    Arguments:
        query (str): any valid SQL query that returns tabular data

    Returns:
        pd.DataFrame: a pandas dataframe with all rows/columns from the query
    """

    engine = sqlalchemy.create_engine(self.uri)

    df = pd.read_sql(query, engine)

    engine.dispose()

    return df



from pg_data_etl import helpers


def export_table_to_another_db(self, table_to_copy: str, target_db) -> None:
    """
    - Pipe data directly from a pg_dump of one DB into another using psql

    Arguments:
        table_to_copy (str): name of table that you want to copy
        target_db (Database): database where you want the data copied to

    Returns:
        None: although it makes a copy of the table inside `target_db`
    """

    # If the table_to_copy has a schema, ensure that the schema also exists in the target db
    if "." in table_to_copy:
        schema = table_to_copy.split(".")[0]
        target_db.schema_add(schema)

    pg_dump = self.cmd.pg_dump
    command = f"{pg_dump} --no-owner --no-acl -t {table_to_copy} {self.uri} | psql {target_db.uri}"

    print(command)
    helpers.run_command_in_shell(command)

    self.gis_table_lint_geom_colname(target_db, table_to_copy)

    return None


def export_entire_db_to_another_db(self, target_db) -> None:
    """
    - Copy an entire database to a new database.

    - To get around memory error limitations, this is done in two steps as opposed to a single
    command with a pipe:
        Step 1) Backup the source db to .sql file with pg_dump
        Step 2) Load the .sql file into the target db with psql

    Arguments:
        target_db (Database): new database (that doesn't exist yet) where you want the data

    Returns:
        None: although it makes a full copy the source database in `target_db`
    """

    if target_db.exists():
        target_db_name = target_db.connection_params["db_name"]
        print(f"A database named '{target_db_name}' already exists.")
        print("Use a different name or drop this database first before copying into it.")
        return None

    else:
        sql_filepath = self.dump()

        target_db.admin("CREATE")

        command = f'{target_db.cmd.psql} -f  "{sql_filepath}" {target_db.uri}'
        helpers.run_command_in_shell(command)

        # Ensure that spatial tables have 'geom' instead of 'shape' columns
        for table in target_db.tables(spatial_only=True):
            target_db.gis_table_lint_geom_colname(table)

        # Delete the .sql file from disk
        sql_filepath.unlink()

        return None



from __future__ import annotations
from pathlib import Path

from pg_data_etl import helpers


def load_from_dumpfile(self, filepath: str | Path) -> None:
    """
    - Load a `.sql` file into a new database
    - The database must not already exist

    Arguments:
        filepath (str): full path to the SQL file

    Returns:
        loads the `.sql` file into an empty database
    """

    db_name = self.connection_params["db_name"]

    if self.exists():
        print(f"Database {db_name} already exists. Use a different name.")
    else:
        command = f'"{self.cmd.psql}" -f "{filepath}" {self.uri}'
        helpers.run_command_in_shell(command)

    return None



from __future__ import annotations
from pathlib import Path
import pandas as pd
import sqlalchemy

from pg_data_etl import helpers


def import_file_with_pandas(
    self,
    filepath: Path | str,
    tablename: str,
    pd_read_kwargs: dict = {},
    df_import_kwargs: dict = {"index": False},
) -> None:
    """
    - Import a tabular CSV or XLSX file to postgres
    - Custom arguments can be provided for the reading of the file via `pd_read_kwargs`
    - Custom import arguments can be provided via `df_import_kwargs`

    Arguments:
        filepath (Path | str): Path or string of filepath to the source CSV or XLSX
        tablename (str): name the new table should be given in the database
        pd_read_kwargs (dict): a key/value dict with any special arguments needed to read the file
        df_import_kwargs (dict): a key/value dict with any special arguments needed to write the data to SQL

    Returns:
        creates a new SQL table from the specified file
    """

    # Determine if this is a CSV, XLS, or XLSX and use the appropriate pandas loader
    filepath = Path(filepath)
    suffix = filepath.suffix.lower()

    if suffix == ".csv":
        df = pd.read_csv(filepath, **pd_read_kwargs)
    elif suffix in [".xlsx", ".xls"]:
        df = pd.read_excel(filepath, **pd_read_kwargs)
    else:
        print(
            f"File type: '{suffix}' is not supported. Check the official pandas documentation to see if this is a valid filetype."
        )
        return None

    self.import_dataframe(df, tablename, df_import_kwargs)


def import_dataframe(self, df: pd.DataFrame, tablename: str, df_import_kwargs: dict = {}) -> None:
    """
    - Import an in-memort `pandas.DataFrame` into postgres

    Arguments:
        df (pd.DataFrame): data to load into postgres, as an in-memory dataframe
        tablename (str): name of the new table in SQL
        df_import_kwargs (dict): a key/value dict with any special arguments needed to write the data to SQL

    Returns:
        creates a new SQL table from the provided dataframe
    """

    # Clean up column names
    df = helpers.sanitize_df_for_sql(df)

    # Make sure the schema exists
    schema, tbl = helpers.convert_full_tablename_to_parts(tablename)
    self.schema_add(schema)

    # Write to database
    engine = sqlalchemy.create_engine(self.uri)

    df.to_sql(tbl, engine, schema=schema, **df_import_kwargs)

    engine.dispose()



"""
`pg_data_etl.database`
----------------------
"""

from __future__ import annotations
from pathlib import Path

from pg_data_etl import helpers
from pg_data_etl.settings import configurations


class Database:
    """
    The `Database` class encapsulates the PostgreSQL connection and
    all necessary functionality. It can be created in two ways:

    - Option 1: pass a database connection string (i.e. 'uri')

        ```python
        >>> uri = 'postgresql://postgres:password@localhost:5432/name_of_db'
        >>> db = Database.from_uri(uri)
        ```

    - Option 2: pass a keyword-argument dictionary with the connection parameters

        ```python
        >>> creds = {'db_name': 'my_db', 'un': 'my_username', 'pw'='my_password'}
        >>> db = Database.from_parameters(**creds)
        ```

    - Option 3: use the built-in configuration manager

        ```python
        >>> db = Database.from_config('my_db', 'localhost')
        ```

    ---

    No matter how it was created, you can access the connection parameters and URI:

    - Get the database URI with `db.uri`

        ```python
        >>> db.uri
        'postgresql://postgres:password@localhost:5432/name_of_db'
        ```

    - Get the database connection parameters with `db.connection_params`

        ```python
        >>> db.connection_params
        {'db_name': 'name_of_db', 'host': 'localhost', 'un': 'postgres',
        'pw': 'password', 'port': '5432', 'extras': None}
        ```

    ---

    """

    def __init__(self, **kwargs):
        """
        - Save all initial keyword arguments as private variables
            - i.e. "host" argument becomes accessible as `Database._host`

        - Accept a flexible set of kwargs, depending on whether the instance
        is created `from_parameters()` or `from_uri()`

        """

        self._can_create_schemas = True
        self._init_kwargs = kwargs

        # Save all kwargs as private variables
        for key, value in kwargs.items():
            setattr(self, f"_{key}", value)

        # Set up the CommandPath subclass
        if self._bin_paths:
            self.cmd = helpers.CommandPathManager(**self._bin_paths)
        else:
            self.cmd = helpers.CommandPathManager()

        # Record how the instance was created
        if hasattr(self, "_uri"):
            self.CREATED_BY_URI = True
        else:
            self.CREATED_BY_URI = False

    @property
    def connection_params(self) -> dict:
        """
        - Return a `dict` of the connection parameters
        """
        if self.CREATED_BY_URI:
            return helpers.decode_uri(self._uri)

        else:
            return {
                "db_name": self._db_name,
                "host": self._host,
                "un": self._un,
                "pw": self._pw,
                "port": self._port,
                "extras": self._extras,
            }

    @property
    def uri(self) -> str:
        """
        - Return a connection string
        """
        if self.CREATED_BY_URI:
            return self._uri
        else:
            return helpers.generate_uri(
                db_name=self._db_name,
                host=self._host,
                un=self._un,
                pw=self._pw,
                port=self._port,
                extras=self._extras,
            )

    @property
    def uri_superuser(self, super_db: str = "postgres"):
        """
        - Return a connection string for the super-user / super-database
        - This is only needed to create the database from within Python
        """

        if self.CREATED_BY_URI:
            params = helpers.decode_uri(self._uri)
            params["db_name"] = super_db
            return helpers.generate_uri(**params)

        else:
            return helpers.generate_uri(
                db_name=self._super_db,
                host=self._host,
                un=self._super_un,
                pw=self._super_pw,
                port=self._port,
            )

    # Creation methods
    # ----------------

    @classmethod
    def from_parameters(
        cls,
        db_name: str,
        host: str = "localhost",
        un: str = "postgres",
        pw: str = "",
        port: int = 5432,
        super_db: str = "postgres",
        super_un: str | None = None,
        super_pw: str | None = None,
        extras: str | None = None,
        bin_paths: dict | None = None,
    ) -> Database:
        """
        - Build a `Database` from keyword arguments

        - Super username/password are necessary to create a database. Unless
        explicitly declared, these values are assumed to be the same as the
        provided username/password

        """

        if not super_un:
            super_un = un
        if not super_pw:
            super_pw = pw

        return cls(
            db_name=db_name,
            host=host,
            un=un,
            pw=pw,
            port=port,
            super_un=super_un,
            super_pw=super_pw,
            super_db=super_db,
            extras=extras,
            bin_paths=bin_paths,
        )

    @classmethod
    def from_uri(
        cls,
        uri: str,
        bin_paths: dict | None = None,
    ) -> Database:
        """
        - Build a `Database` through its URI
        """
        return cls(uri=uri, bin_paths=bin_paths)

    @classmethod
    def from_config(
        cls,
        db_name: str,
        config_key: str,
        config_filepath: str | None = None,
        bin_paths: dict | None = None,
    ) -> Database:
        """
        - Build a `Database` with the configuration file support
        """
        if config_filepath:
            config = configurations(filepath=Path(config_filepath))
        else:
            config = configurations()

        return cls.from_parameters(db_name=db_name, bin_paths=bin_paths, **config[config_key])

    # Administration
    # --------------

    from .actions import exists, admin, schema_add, load_from_dumpfile

    # Change Things Within Database
    # -----------------------------

    from .actions import (
        execute,
        table_add_uid_column,
        table_rename_column,
        gis_make_geotable_from_query,
        gis_table_update_spatial_data_projection,
        gis_table_lint_geom_colname,
        gis_table_add_spatial_index,
    )

    # Lists of Content
    # ----------------

    from .actions import tables, schemas, columns, views, report_spatial, projection

    # Get Data Out of Database To Memory
    # ----------------------------------

    from .actions import gdf, df
    from .actions import query_as_list_of_lists as query
    from .actions import (
        query_as_singleton,
        query_as_list_of_singletons,
        query_as_list_of_lists,
    )

    # Get Data Out of Database To File
    # --------------------------------

    from .actions import (
        dump,
        export_gis,
        export_entire_db_to_another_db,
        export_table_to_another_db,
    )

    # Put Files Into Database
    # -----------------------

    from .actions import import_gis, import_file_with_pandas

    # Put In-Memory Data Into Database
    # --------------------------------

    from .actions import import_dataframe, import_geodataframe



from __future__ import annotations

from pg_data_etl import helpers


def gis_table_lint_geom_colname(self, tablename: str) -> None:
    """
    - Rename the geometry column to 'geom' if it comes through as 'shape'
    - This is necessary for any data coming from an ESRI/PostGIS database,
    where the `geom` convention is replaced by ESRI's preference for `shape`

    Arguments:
        tablename (str): name of the spatial table to lint

    Returns:
        None: but updates the geometry column name if it came through as `shape`
    """

    if "shape" in self.columns(tablename):
        self.table_rename_column("shape", "geom", tablename)


def gis_table_add_spatial_index(self, tablename: str) -> None:
    """
    - Add a spatial index on a table's geom column

    Arguments:
        tablename (str): name of the table to perform a spatial index on

    Returns:
        None: although it updates the database to have a spatial index on the `geom` column
    """
    query = f"""
        CREATE INDEX ON {tablename}
        USING GIST (geom);
    """
    self.execute(query)


def gis_table_update_spatial_data_projection(
    self,
    tablename: str,
    old_epsg: int | str,
    new_epsg: int | str,
    geom_type: str,
) -> None:
    """
    - Transform a table in-place from `old_epsg` to `new_epsg`.
    - You can use this with identical old and new epsgs to force an entry into the
    `geometry_columns` table. (Helpful for making geotables directly in the DB via query)

    Arguments:
        tablename (str): name of the spatial table to re-project
        old_epsg (int | str): EPSG code of the original projection
        new_epsg (int | str): EPSG code of the desired new projection
        geom_type (str): PostGIS geometry data type (e.g 'LineString', 'Point', etc.)

    Returns:
        None: but updates the table in-place to the new_epsg
    """

    query = f"""
        ALTER TABLE {tablename}
        ALTER COLUMN geom TYPE geometry({geom_type}, {new_epsg})
        USING ST_Transform(ST_SetSRID(geom, {old_epsg}), {new_epsg});
    """
    self.execute(query)


def gis_make_geotable_from_query(
    self,
    query: str,
    new_table_name: str,
    geom_type: str,
    epsg: int,
    uid_col: str = "uid",
) -> None:
    """
    - Allow the creation of a new table in the db directly via query.

    - This is especially helpful when you're working with a large dataset
    and you want to limit the I/O processing time.

    Arguments:
        query (str): any valid SQL query that returns a spatial table
        new_table_name (str): name of the new table to hold the query output
        geom_type (str): PostGIS geometry data type returned by the query
        epsg (int): EPSG code of the geometry data returned by the query
        uid_col (str): name of the new unique ID column that will be auto-generated (defaults to 'uid')

    Returns:
        None: but generates a new spatial table from the query
    """

    valid_geom_types = [
        "POINT",
        "MULTIPOINT",
        "POLYGON",
        "MULTIPOLYGON",
        "LINESTRING",
        "MULTILINESTRING",
    ]

    if geom_type.upper() not in valid_geom_types:
        for msg in [
            f"Geometry type of {geom_type} is not valid.",
            f"Please use one of the following: {valid_geom_types}",
            "Aborting",
        ]:
            print("\t", msg)
        return None

    else:

        # Make sure the schema exists
        schema, _ = helpers.convert_full_tablename_to_parts(new_table_name)
        self.schema_add(schema)

        query_to_make_table = f"""
            DROP TABLE IF EXISTS {new_table_name};
            CREATE TABLE {new_table_name} AS
            {query}
        """

        self.execute(query_to_make_table)

        self.table_add_uid_column(new_table_name, uid_col=uid_col)
        self.gis_table_add_spatial_index(new_table_name)

        # We're not reprojecting here, but rather forcing an entry for
        # the new geo table into the geometry_columns table
        self.gis_table_update_spatial_data_projection(new_table_name, epsg, epsg, geom_type.upper())



from __future__ import annotations


def decode_uri(uri: str) -> dict:
    """
    - Turn a URI into a dictionary of parameters
    - e.g. `"postgresql://postgres:password@localhost:5432/name_of_db"`
    becomes:
        {'db_name': 'name_of_db',
        'host': 'localhost',
        'un': 'postgres',
        'pw': 'password',
        'port': '5432',
        'extras': None}

    Args:
        uri (str): database connection string

    Returns:
        dictionary with individual parameters
    """

    psql_un_pw, host_port_db = uri.split("@")

    # Get username and password
    _, un_pw = psql_un_pw.split(r"//")
    un, pw = un_pw.split(":")

    # Get host, port, database name
    host, port_db = host_port_db.split(":")
    port, db_plus = port_db.split(r"/")

    if "?" in db_plus:
        db, extras = db_plus.split("?")
    else:
        db = db_plus
        extras = None

    return {
        "db_name": db,
        "host": host,
        "un": un,
        "pw": pw,
        "port": port,
        "extras": extras,
    }


def generate_uri(
    db_name: str,
    host: str = "localhost",
    un: str = "postgres",
    pw: str = "",
    port: int = 5432,
    extras: str | None = None,
) -> str:
    """
    - Turn individual connection parameters into a URI

    Args:
        db_name (str): name of database
        host (str): name of host
        un (str): username
        pw (str): password
        port (int): port
        extras (str): optional arguments at the end of the connection string

    Returns:
        database connection string
    """
    uri = f"postgresql://{un}:{pw}@{host}:{port}/{db_name}"

    if extras:
        uri += f"?{extras}"

    return uri



from __future__ import annotations
from pg_data_etl import helpers


def tables(self, spatial_only: bool = False, schema: str | None = None) -> list:
    """
    - Return a list of tables in the database
    - Set `spatial_only=True` if you only want a list of geotables

    Arguments:
        spatial_only (bool): flag that will filter to spatial tables only if `True`
        schema (str | None): optional name of schema to filter results to

    Returns:
        list: with all tablenames, with each entry formatted as `schema.tablename`
    """
    if spatial_only:
        query = """
            SELECT concat(f_table_schema, '.', f_table_name )
            FROM geometry_columns
        """

        if schema:
            query += f" WHERE f_table_schema = '{schema}'"

    else:
        query = """
            SELECT concat(table_schema, '.', table_name )
            FROM information_schema.tables
            WHERE table_schema NOT IN ('pg_catalog', 'information_schema')
        """

        if schema:
            query += f" AND table_schema = '{schema}'"

    return self.query_as_list_of_singletons(query)


def schemas(self) -> list:
    """
    - Get a list of all schemas in the db

    Returns:
        list: with all schema names within the database
    """

    query = """
        SELECT schema_name
        FROM information_schema.schemata;
    """

    return self.query_as_list_of_singletons(query)


def views(self) -> list:
    """
    - Get a list of all views in the db

    Returns:
        list: with the names of all views inside the database
    """

    query = """
        select concat(table_schema, '.', table_name)
        from information_schema.views
        where table_schema not in ('information_schema', 'pg_catalog')
    """

    return self.query_as_list_of_singletons(query)


def columns(self, tablename: str) -> list:
    """
    - Get a list of all column names in a given table

    Arguments:
        tablename (str): name of the table you're interested in

    Returns:
        list: with each entry being a column name within the table
    """

    schema, tbl = helpers.convert_full_tablename_to_parts(tablename)

    query = f"""
        SELECT DISTINCT column_name
        FROM information_schema.columns
        WHERE
            table_name = '{tbl}'
            AND
            table_schema = '{schema}';
    """

    return self.query_as_list_of_singletons(query)



from __future__ import annotations
from datetime import datetime


def timestamp_for_filepath(dt: datetime | None = None) -> str:
    """
    - Make a datetime string formatted like: 'on_2021_02_09_at_09_29_58'

    Arguments:
        dt (datetime | None): Use the provided `datetime`, or the current datetime if None

    Returns:
        str: text representation of date and time to use as suffix for database backup
    """
    if not dt:
        dt = datetime.now()

    return dt.strftime("on_%Y_%m_%d_at_%H_%M_%S")



import psycopg2


def execute(self, query: str) -> None:
    """
    - Use psycopg2 to execute a query & commit it to the database

    Arguments:
        query (str): any valid SQL code that changes the database

    Returns:
        None: although the database is updated in-place with whatever is in the query
    """

    connection = psycopg2.connect(self.uri)
    cursor = connection.cursor()

    cursor.execute(query)

    cursor.close()
    connection.commit()
    connection.close()

    return None



from __future__ import annotations
import subprocess
from pathlib import Path


def run_command_in_shell(command: str) -> str:
    """
    - Use subprocess to execute a command in a shell

    Arguments:
        command (str)

    Returns:
        str: output from the command that was run
    """

    process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)

    output, err = process.communicate()
    print(output.decode("utf-8"))

    if err:
        print(err.decode("utf-8"))

    process.kill()

    return output


class CommandPathManager:
    """
    The CommandPathManager class allows customization of all necessary
    CLI tools that this Python package builds on top of.

    It starts with the assumption that everything is on path, and that the
    default version of the command is the one you want to use.

    In some cases, you may have a `psql` installed in your Python environment
    that's incompatible with the database cluster you're connecting to. This class
    exists for this use case, where any and all commands can be configured to execute
    via an explicitly declared full path.

    This means you can alias 'psql' to something like '/my/custom/path/to/psql'

    Examples:
        >>> # use default behavior
        >>> cmd = CommandPathManager()
        >>> cmd.psql
        'psql'

        >>> # use custom path to psql
        >>> cmd = CommandPathManager(psql='/custom/path/to')
        >>> cmd.psql
        '/custom/path/to/psql'

    """

    def __init__(self, **kwargs):
        """
        - Define a dictionary of `bin_paths` for `psql` and `tippecanoe` as `None`
        - Assign any other bin paths provided at runtime
        """
        self.bin_paths = {"psql": None, "tippecanoe": None}

        self.set_bin_path(**kwargs)

    @property
    def configured_commands(self) -> list:
        """
        - Return a list of all command IDs that have been set
        """
        return list(self.bin_paths.keys())

    def _add_bin_path_if_exists(self, cmd: str, bin_id: str) -> str | None:
        """
        - Generate a full path for a `cmd` if the `bin_id` is configured.
        - Otherwise just return the `cmd` without a path, under the assumption that
        it is already on the user's system path

        Args:
            cmd (str): the command to be run from a terminal
            bin_id (str): the bin ID value to inherit the path from

        Returns:
            the full path or shorthand command
        """

        if bin_id in self.configured_commands:
            if self.bin_paths[bin_id]:
                return str(Path(self.bin_paths[bin_id]) / cmd)
            else:
                return cmd
        else:
            print(f"{bin_id=} is not an option. Use one of: {self.configured_commands}")
            return None

    def set_bin_path(self, **kwargs) -> None:
        """
        - Assign bin path to command using `cmd=bin_path` syntax
        """
        for k, v in kwargs.items():
            self.bin_paths[k] = v

    @property
    def psql(self):
        """
        - `psql` connects to a postgres db
        """
        return self._add_bin_path_if_exists("psql", bin_id="psql")

    @property
    def pg_dump(self):
        """
        - `pg_dump` exports a table or full database
        """
        return self._add_bin_path_if_exists("pg_dump", bin_id="psql")

    @property
    def shp2pgsql(self):
        """
        - `shp2pgsql` imports shapefiles into postgres
        """
        return self._add_bin_path_if_exists("shp2pgsql", bin_id="psql")

    @property
    def pgsql2shp(self):
        """
        - `pgsql2shp` exports spatial postgres data to shapefile
        """
        return self._add_bin_path_if_exists("pgsql2shp", bin_id="psql")

    @property
    def ogr2ogr(self):
        """
        - `ogr2ogr` converts spatial data between a variety of formats
        """
        return self._add_bin_path_if_exists("ogr2ogr", bin_id="psql")

    @property
    def tippecanoe(self):
        """
        - `tippecanoe` converts `.geojson` data into vector tiles
        - This is the only command that is not expected to exist in the `psql` bin
        """
        return self._add_bin_path_if_exists("tippecanoe", bin_id="tippecanoe")
